/*********************************************************************/
/* file: alias.c - funtions related the the alias command            */
/*                             HUT TINTIN                            */
/*                coded by Ossi Tulonen from TinTin III              */
/*                TinTin III coded by peter unold 1992               */
/*********************************************************************/
/*
 * TinTin III code Copyright (c) 1992 Peter Unold
 * All changed code from TinTin III Copyright (c) 1992 1993 1994 Ossi tulonen
 * All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 */
 

#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "structs.h"
#include "action.h"
#include "alias.h"
#include "llist.h"
#include "parse.h"
#include "screen.h"
#include "utility.h"
#include "utils.h"

extern struct master_table_data master_table;
extern unsigned char vars[10][BUFFER_SIZE]; /* the $0, $1, $2,..$9 variables */
extern unsigned char text_start, text_end;
extern unsigned char verbatim_char;

/**********************/
/* the #alias command */
/**********************/

struct session *alias_command(const unsigned char *arg, struct session *ses)
{
  unsigned char left[BUFFER_SIZE], right[BUFFER_SIZE], result[BUFFER_SIZE/2];
  const unsigned char *tmp, *tmp2;
  int found, amount, pos, ret_value;
  struct listnode **myaliases, *ln;

  myaliases = &ses->aliases;
  
  arg=get_arg_stop_spaces(arg, left);
  if(*left) {
    pos = 1;
    while(left[pos])
      pos++;
    pos--;
    while(isspace(left[pos]))
      pos--;
    left[pos + 1] = '\0';
  }
  arg=get_arg_with_spaces(arg, right);
  
  if(!*left) {
    ses->print((const uchar *)"#THESE ALIASES HAS BEEN DEFINED:\n\r");
    show_list(*myaliases, ses);
  }

  else if(*left && !*right) {
    for(found = FALSE,ln = searchnode_list2(*myaliases, left); ln;
                      ln = searchnode_list2(ln->next, left)) {
      found = TRUE;
      ln->show(ses);
    }
    if(!found)
      ses->print((const uchar *)"#THAT ALIAS IS NOT DEFINED.\n\r");
  }
  
  else {
    tmp = left;
    while(*tmp > ' ')
      tmp++;
    while(*tmp == ' ')
      tmp++;
    
    found = FALSE;
    if(*tmp) {    // there is format given.
      amount = search_vars(right, result);
      
      for(pos = 0; pos < amount && !found; pos++) {
	tmp2 = tmp;
	while(*tmp2) {
	  if((ret_value = is_variable(tmp2)) != 0 &&
	     ((ret_value < 0 && *(tmp2 + 1) == result[pos]) ||
	      (ret_value > 0 && *(tmp2 + 3) == result[pos])))
	    break;
	  
	  tmp2++;
	}
	if(!*tmp2)
	  found = TRUE;
      }
    }
    if(found) {
      ses->print((const uchar *)"#Atleast one variable would have been replaced with null string.\n\r");
      ses->print((const uchar *)"#Alias NOT defined.\n\r");
    } else {
      if((ln=searchnode_list(*myaliases, left))!=NULL)
	delete ln;
      ln = new listnode(myaliases, left, right, TYPE_ALIAS, 5);
      if(IS_SET(ses->message_flags, MESSAGE_ALIAS)) {
	ses->print((const uchar *)"#OK. ALIAS DEFINED.\n\r");
      }
    }
  }
  return(ses);
}

/************************/
/* the #unalias command */
/************************/

struct session *unalias_command(const unsigned char *arg, struct session *ses)
{
  unsigned char left[BUFFER_SIZE];
  int found;
  struct listnode **myaliases, *ln, *next;

  myaliases = &ses->aliases;

  arg = get_arg_stop_spaces(arg, left);
  
  for(found = FALSE,ln = searchnode_list2(*myaliases, left); ln;
      ln = searchnode_list2(next, left)) {
    found = TRUE;
    next = ln->next;
    delete ln;
    if(IS_SET(ses->message_flags, MESSAGE_ALIAS))
      ses->print((const uchar *)"#OK. ALIAS DELETED\n\r");
  }
  if(!found)
    ses->print((const uchar *)"#THAT ALIAS IS NOT DEFINED.\n\r");
  
  return(ses);
}

// *************************************
//   check alias
// *************************************

const unsigned char *check_all_aliases(const unsigned char *command,
				       const unsigned char *text,
				       unsigned char *newcommand,
				       session *ses)
{
  unsigned char alias[BUFFER_SIZE], args[BUFFER_SIZE], *tmp;
  unsigned char buffer[BUFFER_SIZE], *str, *res;
  const unsigned char *endpos;
  int i, j, result, pos, inside;
  struct listnode *ln, *last_ln = 0;
  
  for(i = 0; i < 10; i++)
    *vars[i] = '\0';
  
  for(ln = ses->aliases; ln; ln = ln->next) {
    for(i = 0; *(ln->left + i) > ' '; i++)
      alias[i] = ln->left[i];
    alias[i] = '\0';
    
    if((result = strcmp((const char *)alias, (const char *)command)) == 0) {
      // first one founded
      while(*(ln->left + i) == ' ')
	i++;
      
      if(!*(ln->left + i)) { // no format
	last_ln = ln;
	ln = ln->next;
      }
      
      while(ln) {
	for(i = 0; *(ln->left + i) > ' '; i++)
	  alias[i] = ln->left[i];
	alias[i] = '\0';
	
	while(*(ln->left + i) == ' ')
	  i++;
	
	if(strcmp((const char *)alias, (const char *)command) == 0) {
	  // aliases with same name with format syntax
	  
	  if(*(ln->right) == verbatim_char) {
	    endpos = text + str_cat(args, 0, sizeof(args), text);
	  } else {
	    endpos = get_arg_all(text, args, ses);
	  }
	  
	  pos = str_cat(buffer, 0, sizeof(buffer), (const uchar *)"^");
	  pos = str_cat(buffer, pos, sizeof(buffer), ln->left + i);
	  pos = str_cat(buffer, pos, sizeof(buffer), (const uchar *)"^");
	  
	  i = 0;
	  j = 0;
	  while(*(args + i)) {
	    if(*(args + i) == '\\') {
	      i++;
	      if(*(args + i))
		*(args + j++) = *(args + i++);
	    } else
	      *(args + j++) = *(args + i++);
	  }
	  *(args + j) = '\0';
	  
	  if(check_one_action(args, buffer)) {
	    substitute_vars(ln->right, newcommand, BUFFER_SIZE, TRUE, ses,
			    FALSE);
	    return(endpos);
	  }
	  ln = ln->next;
	} else
	  break;
      }
      if(last_ln) {
	if(*(last_ln->right) == verbatim_char) {
	  endpos = text + str_cat(buffer, 0, sizeof(buffer), text);
	} else {
	  endpos = get_arg_all(text, buffer, ses);
	}
	substitute_vars(buffer, args, sizeof(args), FALSE, ses, FALSE);
	
	tmp = vars[0];
	i = 0;
	while(*(args + i)) {
	  if(*(args + i) == '\\') {
	    i++;
	    if(*(args + i))
	      *tmp++ = *(args + i++);
	  } else
	    *tmp++ = *(args + i++);
	}
	*tmp = '\0';
	
	for(i = 1, str = args; i < 10; i++) {
	  res = vars[i];
	  
	  if(*str == '\"') {
	    str++;
	    while(*str && *str != '\"')
	      *res++ = *str++;
	    
	    if(*str == '\"')
	      str++;
	    
	  } else if(*str == text_start) {
	    str++;
	    inside = 1;
	    
	    while(*str) {
	      if(*str == text_start) {
		inside++;
	      } else if(*str == text_end) {
		inside--;
		if(!inside)
		  break;
	      }
	      *res++ = *str++;
	    }
	    if(*str == text_end)
	      str++;
	    
	  } else if(*str == '\\') {
	    str++;
	    if(*str)
	      *res++ = *str++;
	    
	  } else {
	    while(*str && *str != ' ')
	      *res++ = *str++;
	  }
	  *res = '\0';
	  
	  while(isspace(*str))
	    str++;
	}
	
	pos = substitute_vars(last_ln->right, newcommand, BUFFER_SIZE,
			      TRUE, ses, FALSE);
	
	if(!strcmp((const char *)last_ln->right, (const char *)newcommand) &&
	   *vars[0]) {
	  pos = str_cat(newcommand, pos, BUFFER_SIZE, (const uchar *)" ");
	  pos = str_cat(newcommand, pos, BUFFER_SIZE, vars[0]);
	}
	
	return(endpos);
      }
      return(NULL);
    } if(result > 0)
      return NULL;            // none alias with that name
  }
  return NULL;
}
