/*********************************************************************/
/* File: main.c - main module - signal setup/shutdown etc            */
/*                             TINTIN III                            */
/*          (T)he K(I)cki(N) (T)ickin D(I)kumud Clie(N)t             */
/*                     coded by peter unold 1992                     */
/*********************************************************************/
/*
 *
 * HISTORY
 *
 * $Log: main.cc,v $
// Revision 1.21  1994/11/26  06:20:40  oat
// v1.1
//
// Revision 1.20  1994/11/10  19:16:21  oat
// added quietmode.
//
// Revision 1.19  1994/11/08  23:21:00  oat
// ht v1.0 pl1
//
// Revision 1.18  1994/11/08  17:06:03  oat
// ht v1.0
//
// Revision 1.17  1994/10/07  15:03:23  oat
// Changed start text.
//
// Revision 1.16  1994/10/04  14:01:22  oat
// ht v 1.0 beta
//
// Revision 1.15  1994/09/28  05:20:39  oat
// ht V0.9_27
//
// Revision 1.14  1994/09/17  23:59:28  oat
// v 0.9_26
//
// Revision 1.13  1994/09/17  12:56:11  oat
// v 0.9_25
//
// Revision 1.12  1994/09/14  23:16:17  oat
// *** empty log message ***
//
// Revision 1.11  1994/09/08  18:09:23  oat
// *** empty log message ***
//
// Revision 1.10  1994/08/31  17:55:02  oat
// "\n" changed to be "\n\r"
//
// Revision 1.9  1994/08/30  19:49:12  oat
// changed version number.
//
// Revision 1.8  1994/08/30  13:55:40  oat
// *** empty log message ***
//
// Revision 1.7  1994/08/17  02:33:55  oat
// *** empty log message ***
//
// Revision 1.6  1994/08/12  08:53:18  oat
// old version before new screen history system ;)
//
// Revision 1.5  1994/07/31  00:02:42  oat
// now multi wondow screen should work.
// (so lets see how many bug fixes version come ;)
//
// Revision 1.4  1994/07/28  01:30:38  oat
// chars marked signed or unsigned
//
// Revision 1.3  1994/07/27  16:51:37  oat
// bug fixes.
//
// Revision 1.2  1994/07/27  14:04:15  oat
// Bugs fixes.
//
// Revision 1.1  1994/07/26  17:18:56  oat
// Initial revision
//
 *
 */


#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#ifndef NSTDLIBS
  #include <stdlib.h>
  #include <unistd.h>
#endif
#ifdef OSF1v3
typedef int clockid_t;
#endif

#include <sys/time.h>
#ifdef TERMIOS
#  ifdef SUN
extern "C" {
#  endif
#  include <termios.h>
#  ifdef SUN
}
#  endif
  #include <sys/ioctl.h>
#else
  #include <sys/ioctl.h>
  #ifdef DIRTY_REDEFINE
    #undef TIOCGETP
    #define TIOCGETP 0x40067408
  #endif
  #include <sgtty.h>
#endif
#include <assert.h>

#ifdef NTYPES
  #include "types.h"
#endif

#ifndef sigmask
#  define sigmask(sig) (1 << ((sig) - 1))
#endif

#include "structs.h"
#include "files.h"
#include "history.h"
#include "llist.h"
#include "main.h"
#include "misc.h"
#include "net.h"
#include "parse.h"
#include "screen.h"
#include "session.h"
#include "ticks.h"
#include "utility.h"
#include "utils.h"

#ifndef BADSIG
#  define BADSIG (void (*)(...))-1
#endif

#define CTRLSPACE (const uchar *)"\000"
#define CTRLA     (const uchar *)"\001"
#define CTRLB     (const uchar *)"\002"
#define CTRLC     (const uchar *)"\003"
#define CTRLD     (const uchar *)"\004"
#define CTRLE     (const uchar *)"\005"
#define CTRLF     (const uchar *)"\006"
#define CTRLG     (const uchar *)"\007"
#define CTRLH     (const uchar *)"\010"   /* backspace */
#define CTRLI     (const uchar *)"\011"   /* tab       */
/* #define CTRLJ 10  return */
#define CTRLK     (const uchar *)"\013"
#define CTRLL     (const uchar *)"\014"
/* #define CTRLM 10  return */
#define CTRLN     (const uchar *)"\016"
#define CTRLO     (const uchar *)"\017"
#define CTRLP     (const uchar *)"\020"
#define CTRLQ     (const uchar *)"\021"
#define CTRLR     (const uchar *)"\022"
#define CTRLS     (const uchar *)"\023"
#define CTRLT     (const uchar *)"\024"
#define CTRLU     (const uchar *)"\025"
#define CTRLV     (const uchar *)"\026"
#define CTRLW     (const uchar *)"\027"
#define CTRLX     (const uchar *)"\030"
#define CTRLY     (const uchar *)"\031"
#define CTRLZ     (const uchar *)"\032"
#define DEL       (const uchar *)"\177"
#define ACSI -101
#define INPUTBUF master_table.inputbuf
#define INPUTPOS master_table.inputbuf_pos
#define CURSOR   master_table.cursor_pos
#define TIMESPERSEC 10

/*************** globals ******************/
int speedwalk    = DEFAULT_SPEEDWALK;
int sessionsstarted;
int default_chars_per_line = DEFAULT_CHARS_PER_LINE;
int default_lines_in_scr = DEFAULT_LINES_IN_SCR;
int signalmask = sigmask(SIGALRM) | sigmask(SIGWINCH) | sigmask(SIGPROF);
int quietmode = FALSE;
struct master_table_data master_table;
unsigned char vars[10][BUFFER_SIZE]; /* the $0, $1, $2,....$9 variables */
unsigned char tintin_char = DEFAULT_TINTIN_CHAR;
unsigned char tintin_retchar = DEFAULT_RETURN_CHAR;
unsigned char text_start = DEFAULT_TEX_START_CHAR;
unsigned char text_end = DEFAULT_TEX_END_CHAR;
unsigned char verbatim_char = DEFAULT_VERBATIM_CHAR;

#ifdef TERMIOS
  struct termios original_state;
#else
  struct tchars original_chars;
  struct sgttyb original_state;
#endif

const uchar *version_num = (const uchar *)"1.1";

class key_data {
public:
  key_data();
  ~key_data();
  
  struct key_data *keys[256];
  void (*func)(int, window *);
  static int amount;
};

int key_data::amount = 0;

key_data::key_data()
{
  int i;
  
  for(i = 0; i < 256; i++)
    keys[i] = 0;
  
  func = 0;
  
  key_data::amount++;
}

key_data::~key_data()
{
  int i;
  
  for(i = 0; i < 256; i++) {
    if(keys[i]) {
      delete keys[i];
    }
  }
  
  key_data::amount--;
}

struct key_func_table {
  key_data *key;
  void (*func)(int, window *);
  int always;
} keys[256];

/************ externs *************/
extern int ticker_interrupted;
extern int window_interrupted;
extern const unsigned char *comm_names[];

extern const uchar *CLEARSCREEN;
extern const uchar *AUTOWARPON;
extern const uchar *RESETCOLOR;
extern const uchar *MAXSCROLLSIZE;

/* ************************************************
   functions for keyboard reading and initilization
   ************************************************ */

void insert_char(unsigned char letter, int prompt_len, window *activ)
{
  unsigned char *tmpstr;
  int i, pos, lines;
  const position *cursor;
  
  if(letter >= ' ') {
    tmpstr = INPUTBUF + CURSOR;
    for(i = INPUTPOS; i > CURSOR; i--)
      *(INPUTBUF + i) = *(INPUTBUF + i - 1);
    
    *(INPUTBUF + CURSOR) = letter;
    INPUTPOS++;
    CURSOR++;
    *(INPUTBUF + INPUTPOS) = '\0';
    
    if(activ->get_hist_on()) {
      if(activ->hist_search(INPUTBUF, 0))
	activ->print_win_hist_line(FALSE);
    } else {
      if(activ->get_active_ses()->term_echoing) {
	activ->print(tmpstr);
	
	if(INPUTPOS > CURSOR) {
	  lines = ((INPUTPOS + prompt_len) / window::get_cols()) -
	    ((CURSOR + prompt_len) / window::get_cols());
	  pos = (CURSOR + prompt_len) % window::get_cols();
	  cursor = activ->get_cursor();
	  activ->move(cursor->row - lines, pos + 1);
	}
      }
    }
  } else {
    activ->bell();
  }
}

void func_home(int prompt_len, window *activ)
{
  int lines;
  const position *cursor;
  
  if(activ->get_hist_on()) {
    activ->hist_home();
  } else {
    if(CURSOR) {
      lines = (prompt_len + CURSOR) / window::get_cols();
      cursor = activ->get_cursor();
      activ->move(cursor->row - lines, prompt_len + 1);
      CURSOR = 0;
    }
  }
}

void func_end(int prompt_len, window *activ)
{
  int pos, lines;
  const position *cursor;
  
  if(activ->get_hist_on()) {
    activ->hist_end();
  } else {
    if(CURSOR != INPUTPOS) {
      lines = ((INPUTPOS + prompt_len) / window::get_cols()) -
	((CURSOR + prompt_len) / window::get_cols());
      pos = (INPUTPOS + prompt_len) % window::get_cols();
      cursor = activ->get_cursor();
      activ->move(cursor->row + lines, pos + 1);
      CURSOR = INPUTPOS;
    } else {
      activ->bell();
    }
  }
}

void func_pageup(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->hist_pageup();
  }
}

void func_pagedown(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->hist_pagedown();
  }
}

void func_arrowup(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->hist_backward();
  } else {
    if(master_table.history_pos < HISTORY_SIZE &&
       master_table.history[master_table.history_pos]) {
      activ->remove_user_text();
      
      INPUTPOS = str_cat(INPUTBUF, 0, sizeof(INPUTBUF),
			 master_table.history[master_table.history_pos]);
      CURSOR = INPUTPOS;
      activ->print_prompt(FALSE);
      master_table.history_pos++;
    }
  }
}

void func_arrowdown(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->hist_forward();
  } else {
    if(master_table.history_pos > 1) {
      activ->remove_user_text();
      INPUTPOS = str_cat(INPUTBUF, 0, sizeof(INPUTBUF),
			 master_table.history[master_table.history_pos - 2]);
      CURSOR = INPUTPOS;
      activ->print_prompt(FALSE);
      master_table.history_pos--;
    } else if(master_table.history_pos == 1) {
      activ->remove_user_text();
      *INPUTBUF = '\0';
      INPUTPOS = 0;
      CURSOR = 0;
      activ->print_prompt(FALSE);
      master_table.history_pos--;
    }
  }
}

void func_arrowright(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->hist_moveforward();
  } else {
    if(CURSOR < INPUTPOS) {
      CURSOR++;
      activ->moveforward();
    } else
      activ->bell();
  }
}

void func_arrowleft(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->hist_movebackward();
  } else {
    if(CURSOR > 0) {
      CURSOR--;
      activ->movebackward();
    } else
      activ->bell();
  }
}

void func_delete_char(int prompt_len, window *activ)
{
  unsigned char *tmpstr;
  int i;
  position cursor;
  
  if(CURSOR != INPUTPOS) {
    INPUTPOS--;
    for(i = CURSOR, tmpstr = INPUTBUF + CURSOR; i <= INPUTPOS; i++) {
      *(INPUTBUF + i) = *(INPUTBUF + i + 1);
    }
    *(INPUTBUF + INPUTPOS) = '\0';
    if(activ->get_hist_on()) {
      activ->print_win_hist_line(FALSE);
    } else {
      cursor = *activ->get_cursor();
      activ->print(tmpstr);
      activ->print((const uchar *)" ");
      activ->move(&cursor);
    }
  } else {
    activ->bell();
  }
}

void func_delete_word(int prompt_len, window *activ)
{
  unsigned char *tmpstr, buf[BUFFER_SIZE];
  int i, len;
  position cursor;
  
  if(activ->get_hist_on()) {
    // search current or previos word begin
    while(CURSOR > 0 &&
	  (!*(INPUTBUF + CURSOR) || isspace(*(INPUTBUF + CURSOR))))
      CURSOR--;
    
    while(CURSOR > 0 && *(INPUTBUF + CURSOR - 1) > ' ')
      CURSOR--;
  } else {
    cursor = *activ->get_cursor();
    
    // search current or previos word begin
    while(CURSOR > 0 &&
	  (!*(INPUTBUF + CURSOR) || isspace(*(INPUTBUF + CURSOR)))) {
      CURSOR--;
      
      if(!(--cursor.col)) {
	cursor.col = window::get_cols();
	if(cursor.row > activ->get_scroll()->start)
	  cursor.row--;
      }
    }
    
    while(CURSOR > 0 && *(INPUTBUF + CURSOR - 1) > ' ') {
      CURSOR--;
      
      if(!(--cursor.col)) {
	cursor.col = window::get_cols();
	if(cursor.row > activ->get_scroll()->start)
	  cursor.row--;
      }
    }
    
    activ->move(&cursor);
  }
  for(len = 0;
      CURSOR + len != INPUTPOS && *(INPUTBUF + CURSOR + len) > ' ';
      len++)
    ;
  
  // all spaces after word removed too
  while(CURSOR + len < INPUTPOS && isspace(*(INPUTBUF + CURSOR + len)))
    len++;
  
  if(len > 0) {
    INPUTPOS -= len;
    
    for(i = CURSOR, tmpstr = INPUTBUF + CURSOR; i <= INPUTPOS; i++) {
      *(INPUTBUF + i) = *(INPUTBUF + i + len);
    }
    *(INPUTBUF + INPUTPOS) = '\0';
    
    if(activ->get_hist_on()) {
      activ->print_win_hist_line(FALSE);
    } else {
      for(i = 0; i < len; i++)
	buf[i] = ' ';
      buf[i] = '\0';
      
      activ->print(tmpstr);
      activ->print(buf);
      activ->move(&cursor);
    }
  } else {
    activ->bell();
  }
}

void func_cut_line(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    if(INPUTPOS) {
      *INPUTBUF = '\0';
      INPUTPOS = 0;
      CURSOR   = 0;
      activ->print_win_hist_line(FALSE);
    }
  } else {
    INPUTPOS = CURSOR;
    *(INPUTBUF + INPUTPOS) = '\0';
    activ->clearend();
  }
}

void func_redraw_screen(int prompt_len, window *activ)
{
  redraw_screen();
}

void func_return(int prompt_len, window *activ)
{
  session *ses;

  static uchar inputbuf2[BUFFER_SIZE] = "";
  static int inside = 0;
  
  if(activ->get_hist_on()) {
    activ->hist_search(INPUTBUF);
  } else {
    ses = activ->get_active_ses();
    
    if(ses->term_echoing) {
      do_history(INPUTBUF, activ);
    }
    
    if(add_str(inputbuf2, master_table.inputbuf2_pos, sizeof(inputbuf2),
	       inside, INPUTBUF, FALSE, ses) &&
       *INPUTBUF) {
      INPUTPOS = 0;
      CURSOR   = 0;
      *INPUTBUF = '\0';
    } else {
      INPUTPOS = 0;
      CURSOR   = 0;
      *INPUTBUF = '\0';
      
      master_table.inputbuf2_pos = 0;
      
      master_table.ucommand = TRUE;
      parse_input(inputbuf2, ses);
      master_table.ucommand = FALSE;
      *inputbuf2 = '\0';
      inside = 0;
    }
  }
}

void func_backspace(int prompt_len, window *activ)
{
  unsigned char *tmpstr;
  int i;
  struct position cursor;

  if(CURSOR) {
    INPUTPOS--;
    CURSOR--;
    for(i = CURSOR, tmpstr = INPUTBUF + CURSOR; i <= INPUTPOS; i++) {
      *(INPUTBUF + i) = *(INPUTBUF + i + 1);
    }
    *(INPUTBUF + INPUTPOS) = '\0';
    
    if(activ->get_hist_on()) {
      activ->print_win_hist_line(FALSE);
    } else {
      activ->movebackward();
      cursor = *activ->get_cursor();
      
      activ->print(tmpstr);
      activ->print((const uchar *)" ");
      activ->move(&cursor);
    }
  } else {
    activ->bell();
  }
}

void func_wordfill(int prompt_len, window *activ)
{
  unsigned char *tmpstr, tmp[BUFFER_SIZE];
  const unsigned char *f_word, *s_word;
  int f_word_pos, s_word_pos, i, j, pos, found;
  session *ses;

  ses = activ->get_active_ses();
  
  if(CURSOR) {
    for(i = CURSOR; i > 0 && ownisalnum(*(INPUTBUF + i - 1)); i--)
      ;
    if(i > 0 && *(INPUTBUF + i - 1) == tintin_char)
      i--;

    for(tmpstr = tmp; i < CURSOR; i++)
      *tmpstr++ = *(INPUTBUF + i);
    *tmpstr = '\0';
  } else {
    *tmp = '\0';
  }
  
  f_word_pos = s_word_pos = -1;
  for(i = 0; i < WORD_LIST_SIZE && ses->word_list[i]; i++)
    if(is_abrev(ses->word_list[i], tmp)) {
      f_word_pos = i;
      break;
    }
  if(f_word_pos >= 0) {
    for(; i < WORD_LIST_SIZE && ses->word_list[i]; i++)
      if(!is_abrev(ses->word_list[i], tmp)) {
	s_word_pos = i - 1;
	break;
      }
    if(s_word_pos < 0)
      s_word_pos = i - 1;
    
    f_word = ses->word_list[f_word_pos];
    s_word = ses->word_list[s_word_pos];
    
    for(pos = strlen((const char *)tmp), j = 1;
	*(f_word + pos) && 
	*(f_word + pos) == *(s_word + pos);
	pos++, j++)
      insert_char(*(f_word + pos), prompt_len, activ);
    
    if(!*(f_word + pos) && !*(s_word + pos)) {
      insert_char(' ', prompt_len, activ);
    } else {
      if(!activ->get_hist_on()) {
	pos = str_cat(tmp, 0, sizeof(tmp), (const uchar *)"\n\r#Words:");
	for(i = f_word_pos; i <= s_word_pos; i++) {
	  if((i - f_word_pos) % 5 == 4)
	    pos = str_cat(tmp, pos, sizeof(tmp), (const uchar *)"\n\r#      ");
	  pos = str_cat(tmp, pos, sizeof(tmp), (const uchar *)" '");
	  pos = str_cat(tmp, pos, sizeof(tmp), ses->word_list[i]);
	  pos = str_cat(tmp, pos, sizeof(tmp), (const uchar *)"'");
	}
	pos = str_cat(tmp, pos, sizeof(tmp), (const uchar *)"\n\r");
	ses->print(tmp);
	ses->get_output()->set_ptexrem();
      }
    }
  } else {
    if((f_word = ses->search_hist_word(tmp, pos))) {
      for(i = strlen((const char *)tmp);
	  ownisalnum(*(f_word + i)) && i < pos;
	  i++)
	insert_char(*(f_word + i), prompt_len, activ);
      if(i < pos) {
	insert_char((unsigned char)' ', prompt_len, activ);
      }
    } else {
      found = FALSE;
      if(*tmp == tintin_char && *(tmp + 1)) {
	for(i = 0; *comm_names[i] != '\n'; i++) {
	  if(is_abrev(comm_names[i], tmp + 1)) {
	    for(pos = strlen((const char *)tmp + 1);
		*(comm_names[i] + pos);
		pos++)
	      insert_char(*(comm_names[i] + pos), prompt_len, activ);
	    insert_char((unsigned char)' ', prompt_len, activ);
	    found = TRUE;
	    break;
	  }
	}
      }
      
      if(!found && !activ->get_hist_on()) {
	sprintf((char *)tmp,
		"\n\r#No such word start in word list or this session"
		" history\n\r# (%d lines checked)\n\r",
		ses->get_output()->get_size()->end - 
		ses->get_output()->get_size()->start + 1);
	ses->print(tmp);
	ses->get_output()->set_ptexrem();
      }
    }
  }
}

void func_system_comm(int prompt_len, window *activ)
{
  unsigned char tmp[BUFFER_SIZE];
  int pos;
  
  if(activ->get_hist_on()) {
    activ->bell();
  } else {
    pos = str_cat(tmp, 0, sizeof(tmp), INPUTBUF);
    *INPUTBUF = '\0';
    INPUTPOS = 0;
    CURSOR   = 0;
    if(activ->get_active_ses()->term_echoing) {
      do_history(tmp, activ);
    }
    system_command(tmp, activ->get_active_ses());
  }
}

void func_screen_hist(int prompt_len, window *activ)
{
  session *ses;

  if(activ->get_hist_on()) {
    activ->bell();
  } else {
    ses = activ->get_active_ses();
    ses->get_output()->start_win_history(ses);
  }
}

void func_quit(int prompt_len, window *activ)
{
  exit_tintin(0);
}

void func_clear(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->bell();
  } else {
    activ->clear();
    activ->print_prompt(TRUE, FALSE);
  }
}

void func_nextwindow(int prompt_len, window *activ)
{
  activ->next_window_active();
}

void func_exit_win_hist(int prompt_len, window *activ)
{
  if(activ->get_hist_on())
    activ->exit_win_history();
  else
    activ->bell();
}

void func_mark(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->hist_set_mark();
  }
}

void func_copy_clip(int prompt_len, window *activ)
{
  if(activ->get_hist_on()) {
    activ->hist_copy_clip();
  }
}

void func_insert_clip(int prompt_len, window *activ)
{
  unsigned char *tmp;
  session *ses;
  
  ses = activ->get_active_ses();
  
  if(master_table.clip) {
    for(tmp = master_table.clip; *tmp; tmp++)
      insert_char(*tmp, prompt_len, activ);
  } else {
    if(!activ->get_hist_on()) {
      ses->print((const uchar *)"#Nothing in clip buffer.\n\r");
      ses->get_output()->set_ptexrem();
    }
  }
}

void add_key_string(const uchar *str, void (*f)(int, window *))
{
  int i, found = FALSE;
  key_data *tmp;
  
  if(!str || !f) {
    return;
  }
  
  if(!*(str + 1)) {
    if(keys[*str].func != f) {
      
      if(keys[*str].key) {
	found = TRUE;
	delete keys[*str].key;
	keys[*str].key = 0;
      }
      
#ifdef DEBUG
      if(keys[*str].func || found)
	fprintf(stderr, "#INFO: replacing old key with new one\n");
#endif
      
      keys[*str].func = f;
    }
  } else {
    if(keys[*str].func) {
      keys[*str].func = 0;
      found = TRUE;
    }
    
    if(!keys[*str].key) {
      keys[*str].key = new key_data;
    }
    tmp = keys[*str].key;
    str++;
    
    while(*str) {
      if(!tmp->keys[*str]) {
	tmp->keys[*str] = new key_data;
      }
      tmp = tmp->keys[*str];
      str++;
    }
    
    for(i = 0; i < 256; i++) {
      if(tmp->keys[i]) {
	delete tmp->keys[i];
	tmp->keys[i] = 0;
	found = TRUE;
      }
    }
#ifdef DEBUG
    if((tmp->func && tmp->func != f) || found)
      fprintf(stderr, "#INFO: replacing old string with new one\n");
#endif
    
    tmp->func = f;
  }
}

void init_key_table(void)
{
  int i;
  
  for(i = 0; i < 256; i++) {
    keys[i].key  = 0;
    keys[i].func = 0;
    keys[i].always = FALSE;
  }
  
  add_key_string(CTRLSPACE, func_mark);
  add_key_string(CTRLA,     func_home);
  add_key_string(CTRLB,     func_arrowleft);
  add_key_string(CTRLC,     func_quit);
  add_key_string(CTRLD,     func_delete_char);
  add_key_string(CTRLE,     func_end);
  add_key_string(CTRLF,     func_arrowright);
  add_key_string(CTRLG,     func_copy_clip);
  add_key_string(CTRLH,     func_backspace);
  add_key_string(CTRLI,     func_wordfill);
  add_key_string(CTRLK,     func_cut_line);
  add_key_string(CTRLL,     func_redraw_screen);
  add_key_string(CTRLN,     func_arrowdown);
  add_key_string(CTRLO,     func_nextwindow);
  keys[*CTRLO].always = TRUE;
  add_key_string(CTRLP,     func_arrowup);
  add_key_string(CTRLV,     func_screen_hist);
  keys[*CTRLV].always = TRUE;
  add_key_string(CTRLW,     func_delete_word);
  add_key_string(CTRLX,     func_exit_win_hist);
  add_key_string(CTRLY,     func_insert_clip);
  add_key_string(CTRLZ,     func_system_comm);
  add_key_string((const uchar *)"\r", func_return);
  add_key_string((const uchar *)"\n", func_return);
  add_key_string(DEL,       func_backspace);

  add_key_string((const uchar *)"\033[1~", func_home);
  add_key_string((const uchar *)"\033[2~", func_insert_clip);
  add_key_string((const uchar *)"\033[4~", func_end);
  add_key_string((const uchar *)"\033[5~", func_pageup);
  add_key_string((const uchar *)"\033[6~", func_pagedown);
  add_key_string((const uchar *)"\033[A",  func_arrowup);
  add_key_string((const uchar *)"\033[B",  func_arrowdown);
  add_key_string((const uchar *)"\033[C",  func_arrowright);
  add_key_string((const uchar *)"\033[D",  func_arrowleft);
  add_key_string((const uchar *)"\033[[A", func_mark);
  add_key_string((const uchar *)"\033[[B", func_copy_clip);
}

const struct key_elem key_functions[] = {
  {(const uchar *)"home",        func_home},
  {(const uchar *)"arrowleft",   func_arrowleft},
  {(const uchar *)"quit",        func_quit},
  {(const uchar *)"delchar",     func_delete_char},
  {(const uchar *)"end",         func_end},
  {(const uchar *)"arrowright",  func_arrowright},
  {(const uchar *)"backspace",   func_backspace},
  {(const uchar *)"wordfill",    func_wordfill},
  {(const uchar *)"cut_line",    func_cut_line},
  {(const uchar *)"redraw_scr",  func_redraw_screen},
  {(const uchar *)"arrowdown",   func_arrowdown},
  {(const uchar *)"nextwin",     func_nextwindow},
  {(const uchar *)"arrowup",     func_arrowup},
  {(const uchar *)"system",      func_system_comm},
  {(const uchar *)"screen_hist", func_screen_hist},
  {(const uchar *)"clear",       func_clear},
  {(const uchar *)"delword",     func_delete_word},
  {(const uchar *)"exitwinhist", func_exit_win_hist},
  {(const uchar *)"pageup",      func_pageup},
  {(const uchar *)"pagedown",    func_pagedown},
  {(const uchar *)"mark",        func_mark},
  {(const uchar *)"copy_clip",   func_copy_clip},
  {(const uchar *)"insert_clip", func_insert_clip},
  {(const uchar *)"\n", 0}
};

unsigned char ctrl_nums['z' - 'a' + 1] = {
  1,  2,  3,  4,  5,  6,  7,  8,  9,  0,
  11, 12,  0, 14, 15, 16, 17, 18, 19, 20,
  21, 22, 23, 24, 25, 26
  };

session *comm_ctrl(const unsigned char *arg, session *ses)
{
  unsigned char letter[BUFFER_SIZE], name[BUFFER_SIZE], *tmp;
  int pos, i, ctrl_num = -1, always = FALSE;
  
  arg = get_arg_stop_spaces(arg, letter);
  tmp = letter;
  if(*tmp == '-') {
    tmp++;
    while(isspace(*tmp))
      tmp++;
    
    while(*tmp) {
      switch(*tmp) {
      case 'a':
	always = TRUE;
	break;
      default:
	sprintf((char *)name, "#Unknown flag '%c'.\n\r", *tmp);
	ses->print(name);
	break;
      }
      tmp++;
      while(isspace(*tmp))
	tmp++;
    }
    arg = get_arg_stop_spaces(arg, letter);
  }

  arg = get_arg_with_spaces(arg, name);
  
  if(*letter && !*(letter + 1)) {
    if(*letter < 'a' ||
       *letter > 'z' ||
       (ctrl_num = ctrl_nums[*letter - 'a']) < 1) {
      ses->print((const uchar *)"#Sorry you can't use that crtl key\n\r");
      return(ses);
    }
    if(*name) {
      for(i = 0; *key_functions[i].name != '\n'; i++) {
	if(strcmp((const char *)name, (const char *)key_functions[i].name) == 0) {
	  keys[ctrl_num].func = key_functions[i].func;
	  keys[ctrl_num].always = always;
	  if(keys[ctrl_num].key) {
	    delete keys[ctrl_num].key;
	    keys[ctrl_num].key = 0;
	  }
	  ses->print((const uchar *)"#Ok.\n\r");
	  return(ses);
	}
      }
    } else {
      if(keys[ctrl_num].func) {
	for(i = 0; *key_functions[i].name != '\n'; i++) {
	  if(key_functions[i].func == keys[ctrl_num].func) {
	    pos = str_cat(name, 0, BUFFER_SIZE,
			  (const uchar *)"#That ctrl key is: ");
	    pos = str_cat(name, pos, BUFFER_SIZE, key_functions[i].name);
	    if(keys[ctrl_num].always)
	      pos = str_cat(name, pos, BUFFER_SIZE,
			    (const uchar *)" (always)");
	    pos = str_cat(name, pos, BUFFER_SIZE,
			  (const uchar *)"\n\r");
	    ses->print(name);
	    return(ses);
	  }
	}
	ses->print((const uchar *)"#That ctrl key is: UNKNOWN FUNCTION ? :/\n\r");
      } else {
	ses->print((const uchar *)"#That ctrl key is: NOTHING.\n\r");
      }
      return(ses);
    }
  }
  pos = str_cat(name, 0, BUFFER_SIZE,
		(const uchar *)"#Next functions are available:\n\r");
  for(i = 0; *key_functions[i].name != '\n'; i++) {
    pos = str_cat(name, pos, BUFFER_SIZE, (const uchar *)"   ");
    pos = str_cat(name, pos, BUFFER_SIZE, key_functions[i].name);
    pos = str_cat(name, pos, BUFFER_SIZE, (const uchar *)"\n\r");
  }
  ses->print(name);
  return(ses);
}

/**************************************************************************/
/* main() - show title - setup signals - init lists - readcoms - tintin() */
/**************************************************************************/
int main(int argc, char *argv[])
{
  unsigned char buf[1000], *filename = 0, *tmp;
  int i, nextisname = FALSE;
#if !defined(HPUX) && !defined(BSD43) && !defined(SUN)
  struct sigaction siga;
#endif

  for(i = 1; i < argc; i++) {
    tmp = (uchar *)argv[i];

    if(*tmp == '-' && !nextisname) {
      tmp++;
      
      while(isspace(*tmp))
	tmp++;
      
      while(*tmp) {
	switch(*tmp) {
	case '-':
	  nextisname = TRUE;
	  break;
	case 'q':
	case 'Q':
	  quietmode = TRUE;
	  break;
	default:
	  printf("Unknown argument %c.\n\r", *tmp);
	  break;
	}
	tmp++;
	while(isspace(*tmp))
	  tmp++;
      }
      
    } else {
      nextisname = FALSE;
      
      if(filename) {
	printf("Usage: %s [-q] [COMMAND FILENAME]\n\r", argv[0]);
	exit(0);
      }
      
      filename = tmp;
    }
  }
  
  for(i = 0; i < HISTORY_SIZE; i++)
    master_table.history[i] = 0;
  master_table.clip = 0;
  master_table.history_pos = 0;
  *master_table.inputbuf = '\0';
  master_table.inputbuf_pos = 0;
  master_table.inputbuf2_pos = 0;
  master_table.cursor_pos = 0;
  master_table.utexrem    = FALSE;
  master_table.ucommand   = FALSE;
  master_table.aliascalls = 0;
  master_table.verbatimmode = FALSE;
  master_table.no_neg_off = FALSE;
  master_table.output_window = 0;
  
  for(i = 0; i < 10; i++) {
    *vars[i] = '\0';
  }

  init_key_table();
  init_termcap();

#ifdef TERMIOS
  tcgetattr(0, &original_state);
#else
  ioctl(0, TIOCGETP, &original_state);
  ioctl(0, TIOCGETC, &original_chars);
#endif
  cbreak_noecho();
  
  window::get_screen_size(); // this will clear screen

  master_table.main_window  = new window;
  master_table.main_session = new session;
  
  sprintf((char *)buf, "Screen size: %dx%d\n\r",
	  window::get_cols(),
	  window::get_rows());
  
  init_comms_table();
  
  if(!quietmode) {
    master_table.main_session->print(buf);
    
    master_table.main_session->print((const uchar *)"###################################################\n\r"
				     "#                    H U T                        #\n\r"
				     "#                 T I N T I N                     #\n\r"
				     "#                                                 #\n\r"
				     "#  (T)he k(I)cki(N) (T)ickin d(I)kumud clie(N)t   #\n\r"
				     "#  Orginal TINTIN III coded by peter unold 1992   #\n\r"
				     "#                  HUT TINTIN                     #\n\r"
				     "# coded by Ossi Tulonen 1992-1994 from TINTIN III #\n\r"
				     "###################################################\n\r");
  }
  
#if !defined(HPUX) && !defined(BSD43) && !defined(SUN)
  bzero((char *)&siga, sizeof(siga));
  siga.sa_flags   = SA_RESTART;
  
  siga.sa_handler = myquitsig;
  if(sigaction(SIGTERM, &siga, 0))
    syserr((const uchar *)"signal SIGTERM");
  
  siga.sa_handler = myquitsig;
  if(sigaction(SIGINT, &siga, 0))
    syserr((const uchar *)"signal SIGINT");
  
  siga.sa_handler = window_func;
  if(sigaction(SIGWINCH, &siga, 0))
    syserr((const uchar *)"signal SIGWINCH");
  
  siga.sa_handler = SIG_IGN;
  if(sigaction(SIGALRM, &siga, 0))
    syserr((const uchar *)"signal SIGALRM");
  if(sigaction(SIGQUIT, &siga, 0))
    syserr((const uchar *)"signal SIGQUIT");
  if(sigaction(SIGTSTP, &siga, 0))
    syserr((const uchar *)"signal SIGTSTP");
  if(sigaction(SIGPIPE, &siga, 0))   // Ignore broken pipe signal
    syserr((const uchar *)"signal SIGPIPE");
#else
  
  if(signal(SIGTERM, myquitsig) == BADSIG)
    syserr((const uchar *)"signal SIGTERM");
  
  if(signal(SIGINT, myquitsig) == BADSIG)
    syserr((const uchar *)"signal SIGINT");
  
  if(signal(SIGALRM, SIG_IGN) == BADSIG)
    syserr((const uchar *)"signal SIGALRM");
  
  if(signal(SIGWINCH, window_func) == BADSIG)
    syserr((const uchar *)"signal SIGWINCH");
  
  if(signal(SIGQUIT, SIG_IGN) == BADSIG)
    syserr((const uchar *)"signal SIGQUIT");
  
  if(signal(SIGTSTP, SIG_IGN) == BADSIG)
    syserr((const uchar *)"signal SIGTSTP");

  if(signal(SIGPIPE, SIG_IGN) == BADSIG)        // Ignore broken pipe signal
    syserr((const uchar *)"signal SIGPIPE");
#endif
  
  read_command(filename, master_table.main_session);
  
  if(quietmode) {
    quietmode = FALSE;
  } else {
    comm_version((unsigned char *)"", master_table.main_session);
  }
  
  window::get_active_win()->set_ptexrem();
  prompt();
  
  tintin();
  exit_tintin(0);
  exit(0);
}

// **************************************************
//  function to ignoring SIG_ARLM & SIG_WINCH on/off
// **************************************************

void sigmask_on()
{
#if !defined(HPUX) && !defined(BSD43) && !defined(SUN)
  struct sigaction siga;
  
  bzero((char *)&siga, sizeof(siga));
  siga.sa_flags   = SA_RESTART;
  siga.sa_handler = SIG_IGN;
  
  if(sigaction(SIGWINCH, &siga, 0))
    syserr((const uchar *)"signal SIGWINCH");
#else
  if(signal(SIGWINCH, SIG_IGN) == BADSIG)
    syserr((const uchar *)"signal SIGWINCH");
#endif
  
  sigsetmask(signalmask);
}

void sigmask_off()
{
#if !defined(HPUX) && !defined(BSD43) && !defined(SUN)
  struct sigaction siga;
  
  bzero((char *)&siga, sizeof(siga));
  siga.sa_flags   = SA_RESTART;

  siga.sa_handler = window_func;
  if(sigaction(SIGWINCH, &siga, 0))
    syserr((const uchar *)"signal SIGWINCH");
#else
  if(signal(SIGWINCH, window_func) == BADSIG)
    syserr((const uchar *)"signal SIGWINCH");
#endif
  
  sigsetmask(0);
}

/***************************/
/* the main loop of tintin */
/***************************/
void tintin(void)
{
  unsigned char buf[BUFFER_SIZE], *timestr;
  int s, pulse, maxnet, sel_res;
#ifdef NUSLEEP
  struct timeval timeout;
#else
  long timeout;
#endif
  struct timeval null_time;
  fd_set input_set, exc_set;
  session *sestmp, *sesresult;
  session *sesptr, *sesnext, *ses;
  struct waitcommdata *tmp, *tmpnext, *last;
  struct listnode *ln;
  static time_t oldtime = 0;
  window *win;
  
  null_time.tv_sec = 0;
  null_time.tv_usec = 0;
#ifdef NUSLEEP
  timeout.tv_sec = 0;
  timeout.tv_usec = 1000000/TIMESPERSEC;
#else
  timeout = 1000000/TIMESPERSEC;    /* 1 sec / TIMESPERSEC */
#endif
  
  pulse = 0;
  s = fileno(stdin);
  while(TRUE) {
    FD_ZERO(&input_set);
    FD_ZERO(&exc_set);
    
    FD_SET(s, &input_set);
    FD_SET(s, &exc_set);
    maxnet = s;
    for(sesptr = session::get_sessions();
	sesptr;
	sesptr = sesptr->get_next()) {
      FD_SET(sesptr->get_socket(), &input_set);
      FD_SET(sesptr->get_socket(), &exc_set);
      if(sesptr->get_socket() > maxnet)
	maxnet = sesptr->get_socket();
    }
    sigsetmask(signalmask);
#ifdef HPUX
    if((sel_res = select(maxnet + 1, (int *)&input_set, (int *)0, 
			 (int *)&exc_set, &null_time)) < 0)
      syserr((const uchar *)"main.cc select");
#else
    if((sel_res = select(maxnet + 1, &input_set, (fd_set *)0,
			 &exc_set , &null_time)) < 0)
      syserr((const uchar *)"main.cc select");
#endif
    sigsetmask(0);
    
    if(sel_res) {
      for(sesptr = session::get_sessions(); sesptr; sesptr = sesnext) {
	sesnext = sesptr->get_next();
	if(FD_ISSET(sesptr->get_socket(), &exc_set)) {
	  /*	  printf("EXC %s", sesptr->name); */
	}
	
	if(FD_ISSET(sesptr->get_socket(), &input_set)) {
	  sesptr->read_net();
	}
      }
      
      if(FD_ISSET(s, &input_set)) {
	read_user();
      }
    }
#ifdef NUSLEEP
#  ifdef HPUX
    select(1, (int *)0, (int *)0, (int *)0, &timeout);
#  else
    select(1, (fd_set *)0, (fd_set *)0, (fd_set *)0, &timeout);
#  endif
#else
    usleep(timeout);
#endif
    if(oldtime != time(0)) {
      oldtime = time(0);
      
      timestr = own_asctime();
      
      for(ses = master_table.main_session; ses; ses = sesnext) {
	sestmp = ses->get_next();
	sesresult = ses;
	
	if(ses->tickstatus) {
	  ses->sec_to_tick = (ses->tick_size -
			      ((oldtime - ses->tick_start_time) % 
			       ses->tick_size));
	  if(ses->sec_to_tick == ses->tick_size) {
	    ses->get_output()->set_ptexrem();
	    ses->print((const uchar *)"#TICK!!!\n\r");
	    ses->get_output()->print_prompt(TRUE);
	  } else if(ses->sec_to_tick == 10) {
	    ses->get_output()->set_ptexrem();
	    ses->print((const uchar *)"#10 SECONDS TO TICK!!!\n\r");
	    ses->get_output()->print_prompt(TRUE);
	  }
	}
	if((ln = searchnode_list(ses->colors, (unsigned char *)"TIME")))
	  delete ln;
	
	if((ln = searchnode_list(ses->variables, (unsigned char *)"TIME"))) {
	  if(ln->right)
	    free(ln->right);
	  ln->right = (uchar *)strdup((const char *)timestr);
	} else
	  ln = new listnode(&ses->variables, (unsigned char *)"TIME", timestr,
			    TYPE_VAR, 5);
	
	for(tmp = ses->waitcomms; tmp; tmp = tmpnext) {
	  tmpnext = tmp->next;
	  tmp->time--;
	  if(!tmp->time) {
	    if(IS_SET(ses->echo_flags, ECHO_WAIT)) {
	      sprintf((char *)buf, "[WAIT: %s]\n\r", tmp->comm);
	      ses->print(buf);
	    }
	    strcpy((char *)buf, (const char *)tmp->comm);
	    free(tmp->comm);
	    if(tmp == ses->waitcomms) {
	      ses->waitcomms = tmp->next;
	      free(tmp);
	    } else {
	      for(last = ses->waitcomms;
		  last && last->next != tmp;
		  last = last->next)
		;
	      if(last)
		last->next = tmp->next;
	      else
		assert(0);
	      free(tmp);
	    }
	    
	    sesresult = parse_input(buf, ses);
	    prompt();
	    if(!sesresult)
	      break;
	  }
	}
	if(!sesresult)
	  sesnext = sestmp;
	else
	  sesnext = ses->get_next();
      }
      for(win = window::get_windows(); win; win = win->get_next()) {
	if((win->get_delay() > 0  &&
	    win->lastupdate + win->get_delay() <= oldtime) ||
	   (win->get_delay() >= 0 &&
	    win->get_lastactive() != window::get_active_win())) {
	  win->print_time(FALSE);
	  win->lastupdate = oldtime;
	}
      }
      win = window::get_active_win();
      if(win->get_cursor()->col != window::get_scrcursor()->col ||
	 win->get_cursor()->row != window::get_scrcursor()->row)
	win->move(win->get_cursor());
      fflush(stdout);
    }
  }
}

/*************************************************************/
/* read text from user and test for alias                    */
/*************************************************************/
  
void read_user(void)
{
  unsigned char buffer[100], letter, *rbuf, tmp[2] = {'\0', '\0'};
  int didget, i = 0, prompt_len, pos;
  window *activ;
  session *ses;
  key_data *key_tmp;

  static unsigned char oldbuf[20] = "";
  
  if(window::get_active_win()->get_hist_clip())
    window::get_active_win()->hist_rem_clip_color();
  
  rbuf = buffer;
  if(*oldbuf) {
    rbuf += str_cat(rbuf, 0, sizeof(buffer), oldbuf);
    *oldbuf = '\0';
  }
  
  if((didget = read(0, rbuf, sizeof(buffer) - 1 - (rbuf - buffer))) < 0)
    syserr((const uchar *)"read from fd 0");
  didget += rbuf - buffer;
  
  while(INPUTPOS < sizeof(INPUTBUF) - 1 && i < didget) {
    
    // printf("\nnum: %d: ", (int)*(buffer + i));
    
    activ = window::get_active_win();
    ses   = activ->get_active_ses();
    prompt_len = ses->get_prompt_len();
    
    letter = *(buffer + i++);
    if(ses->line_mode ||
       ses->get_socket() < 0 ||
       activ->get_hist_on() ||
       INPUTPOS > 0) {

      if(keys[letter].func)
	(keys[letter].func)(prompt_len, activ);
      else if(keys[letter].key) {
	key_tmp = keys[letter].key;
	pos = 0;
	oldbuf[pos++] = letter;

	while(i < didget && key_tmp->keys[*(buffer + i)]) {
	  oldbuf[pos++] = *(buffer + i);
	  key_tmp = key_tmp->keys[*(buffer + i++)];
	}
	if(key_tmp->func) {
	  oldbuf[0] = '\0';
	  (key_tmp->func)(prompt_len, activ);
	} else if(i < didget) {
	  oldbuf[0] = '\0';
	} else {
	  oldbuf[pos] = '\0';
	}
      } else
	insert_char(letter, prompt_len, activ);
    } else {
      if(keys[letter].func &&
	 keys[letter].always)
	(keys[letter].func)(prompt_len, activ);
      else {
	tmp[0] = letter;
	if(write_socket(tmp, ses->get_socket(), 1) < 0) {
	  perror("write socket");
	  delete ses;
	  return;
	}
      }
    }
  }
  if(i < didget) {
    window::get_active_win()->get_active_ses()->print((const uchar *)"#ERROR - You cannot give that long strings.\n\r");
    INPUTPOS = 0;
    CURSOR   = 0;
    *INPUTBUF = '\0';
  }
  prompt();
  fflush(stdout);
}

/* ********************************************************
 *      release all memory, reset screen and exit         *
 ******************************************************** */

void exit_tintin(int status)
{
  int i, oldamount;
  session *sesptr, *sesnext;
  window *win, *nextwin;
  
  for(sesptr = session::get_sessions(); sesptr; sesptr = sesnext) {
    sesnext = sesptr->get_next();
    delete sesptr;
  }

  for(win = window::get_windows(); win; win = nextwin) {
    nextwin = win->get_next();
    delete win;
  }

  for(i = 0; i < HISTORY_SIZE; i++)
    if(master_table.history[i])
      free(master_table.history[i]); 
  
  oldamount = key_data::amount;

  for(i = 0; i < 256; i++) {
    if(keys[i].key) {
      delete keys[i].key;
      keys[i].key = 0;
    }
  }
  printf("%s%s%s", MAXSCROLLSIZE, RESETCOLOR, CLEARSCREEN);
#ifdef DEBUG
  printf("Keytable size was: %d (=%d bytes)\n\r",
	 oldamount, oldamount*sizeof(key_data));
#endif
  printf("Thanks for using this TinTin.\n\r");
  printf("We hope you like it ;)\n\r");
  printf("Have a nice day (or night).\n\r");
  printf("%s", AUTOWARPON);
  nocbreak_echo();
  exit(status);
}

/**********************************************************/
/* Here's where we go when we wanna quit TINTIN FAAAAAAST */
/**********************************************************/
void myquitsig(int i)
{
  exit_tintin(0);
}
